/* unicap
 *
 * Copyright (C) 2004 Arne Caspari ( arne@unicap-imaging.org )
 *
  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 2 of the License, or
  (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#include "config.h"

#if HAVE_AVCODEC
#include <avcodec.h>
#include <avutil.h>
#endif //HAVE_AVCODEC

#include <sys/types.h>
#include <linux/types.h>
#include <glib.h>
#include <string.h>


#include "colorspace.h"
#include "ucil.h"

#if UNICAPGTK_DEBUG
#define DEBUG
#endif
#include "debug.h"

#define SAT(c) if( ( c & (~255) ) != 0 ) { if (c < 0) c = 0; else c = 255; }

#define FOURCC(a,b,c,d) (unsigned int)((((unsigned int)d)<<24)+ \
                        (((unsigned int)c)<<16)+(((unsigned int)b)<<8)+a)

//52474200-0000-0010-8000-00aa00389b71
#define GUID_RGB24 { 0x52, 0x47, 0x42, 0x00, 0x00, 0x00, 0x00, 0x10, \
		     0x80, 0x00, 0x00, 0xaa, 0x00, 0x38, 0x9b, 0x71 }
#define GUID_BGR24 { 0x42, 0x47, 0x52, 0x00, 0x00, 0x00, 0x00, 0x10, \
		     0x80, 0x00, 0x00, 0xaa, 0x00, 0x38, 0x9b, 0x71 }

//52474241-0000-0010-8000-00aa00389b71
#define GUID_RGB32 { 0x52, 0x47, 0x42, 0x41, 0x00, 0x00, 0x00, 0x10, \
                     0x80, 0x00, 0x00, 0xaa, 0x00, 0x38, 0x9b, 0x71 } 

//55595659-0000-0010-8000-00aa00389b71
#define GUID_UYVY  { 0x55, 0x59, 0x56, 0x59, 0x00, 0x00, 0x00, 0x10, \
		     0x80, 0x00, 0x00, 0xaa, 0x00, 0x38, 0x9b, 0x71 }

//59555932-0000-0010-8000-00aa00389b71
#define GUID_YUY2  { 0x59, 0x55, 0x59, 0x32, 0x00, 0x00, 0x00, 0x10, \
		     0x80, 0x00, 0x00, 0xaa, 0x00, 0x38, 0x9b, 0x71 }

//59555600-0000-0010-8000-00aa00389b71
#define GUID_YUV   { 0x59, 0x55, 0x56, 0x00, 0x00, 0x00, 0x00, 0x10, \
		     0x80, 0x00, 0x00, 0xaa, 0x00, 0x38, 0x9b, 0x71 }

//59383030-0000-0010-8000-00aa00389b71
#define GUID_Y800  { 0x59, 0x38, 0x32, 0x30, 0x00, 0x00, 0x00, 0x10, \
		     0x80, 0x00, 0x00, 0xaa, 0x00, 0x38, 0x9b, 0x71 }

//59563132-0000-0010-8000-00aa00389b71
#define GUID_YV21  { 0x59, 0x56, 0x31, 0x32, 0x00, 0x00, 0x00, 0x10, \
		     0x80, 0x00, 0x00, 0xaa, 0x00, 0x38, 0x9b, 0x71 }

//49343230-0000-0010-8000-00aa00389b71
#define GUID_I420  { 0x49, 0x34, 0x32, 0x30, 0x00, 0x00, 0x00, 0x10, \
		     0x80, 0x00, 0x00, 0xaa, 0x00, 0x38, 0x9b, 0x71 }

//49343230-0000-0010-8000-00aa00389b71
#define GUID_I422  { 0x49, 0x34, 0x32, 0x32, 0x00, 0x00, 0x00, 0x10, \
		     0x80, 0x00, 0x00, 0xaa, 0x00, 0x38, 0x9b, 0x71 }


static void grey2uyvy( __u8 *dest, __u8 *source, int width, int height );
static void uyvy2grey( __u8 *dest, __u8 *source, int width, int height );
static void grey2yuy2( __u8 *dest, __u8 *source, int width, int height );
static void yuy22grey( __u8 *dest, __u8 *source, int width, int height );
static void y8002y420( __u8 *dest, __u8 *source, int width, int height );
static void y8002y422( __u8 *dest, __u8 *source, int width, int height );
static void y4202y800( __u8 *dest, __u8 *source, int width, int height );
static void uyvy2yuy2( __u8 *dest, __u8 *source, int width, int height );
static void uyvy2yuv( __u8 *dest, __u8 *source, int width, int height );
static void uyvy2rgb24( __u8 *dest, __u8 *source, int width, int height );
static void yuyv2rgb24( __u8 *dest, __u8 *source, int width, int height );
static void uyvy2rgb32( __u8 *dest, __u8 *source, int width, int height );
static void y4112rgb24( __u8 *dest, __u8 *source, int width, int height );
static void y4112rgb32( __u8 *dest, __u8 *source, int width, int height );
static void y4202rgb24( __u8 *dest, __u8 *source, int width, int height );
static void y4202rgb32( __u8 *dest, __u8 *source, int width, int height );
static void y8002rgb24( __u8 *dest, __u8 *source, int width, int height );
static void y8002rgb32( __u8 *dest, __u8 *source, int width, int height );
static void uyvytoyuv422p( __u8 *dest, __u8 *src, int width, int height );
static void uyvytoyuv420p( __u8 *dest, __u8 *src, int width, int height );
static void yuv420ptouyvy( __u8 *dest, __u8 *src, int width, int height );
static void yuyvtoyuv420p( __u8 *dest, __u8 *src, int width, int height );
static void yuv420ptoyuyv( __u8 *dest, __u8 *src, int width, int height );
static void yuv420ptorgb24( __u8 *dest, __u8 *source, int width, int height );
static void bgr242uyvy( __u8 *dest, __u8 *source, int width, int height );
static void rgb242uyvy( __u8 *dest, __u8 *source, int width, int height );
static void rgb322uyvy( __u8 *dest, __u8 *source, int width, int height );
static void rgb24toyuv420p( __u8 *dest, __u8 *source, int width, int height );
static void bgr24torgb24( __u8 *dest, __u8 *source, int width, int height );
static void rgb32torgb24( __u8 *dest, __u8 *source, int width, int height );
static void yuy2toyuyv( __u8 *dest, __u8 *source, int width, int height );
static void by8touyvy( __u8 *dest, __u8* source, int width, int height );
static void by8torgb24( __u8 *dest, __u8* source, int width, int height );
static void by8torgb32( __u8 *dest, __u8* source, int width, int height );
static void by8tobgr24( __u8 *dest, __u8* source, int width, int height );
static void by8touyvy( __u8 *dest, __u8* source, int width, int height );


static xfm_info_t xfminfo = 
{
   src_fourcc: 0xffffffff,
   dest_fourcc:   0xffffffff,
   func: NULL
};

static xfm_info_t conversions[] =
{
   { 
      src_fourcc: FOURCC('G','R','E','Y'), 
      src_guid: { 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 
		  0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00 }, 
      src_bpp: 8, 
      
      dest_fourcc: FOURCC('U','Y','V','Y'), 
      dest_guid: GUID_UYVY, 
      dest_bpp: 16,
      
      priority: 5, 

      func: grey2uyvy,

      flags: 0
   },
   { 
      src_fourcc: FOURCC('Y','8','0','0'), 
      src_guid: { 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 
		  0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00 }, 
      src_bpp: 8, 
      
      dest_fourcc: FOURCC('U','Y','V','Y'), 
      dest_guid: GUID_UYVY, 
      dest_bpp: 16,
      
      priority: 5, 

      func: grey2uyvy,

      flags: 0
   },
   { 
      src_fourcc: FOURCC('G','R','E','Y'), 
      src_guid: { 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 
		  0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00 }, 
      src_bpp: 8, 
      
      dest_fourcc: FOURCC('U','Y','V','Y'), 
      dest_guid: GUID_UYVY, 
      dest_bpp: 16,
      
      priority: 5, 

      func: grey2uyvy,

      flags: 0
   },
   { 
      src_fourcc: FOURCC('Y','8','0','0'), 
      src_guid: { 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 
		  0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00 }, 
      src_bpp: 8, 
      
      dest_fourcc: FOURCC('I','4','2','0'), 
      dest_guid: GUID_UYVY, 
      dest_bpp: 12,
      
      priority: 5, 

      func: y8002y420,

      flags: 0
   },
   { 
      src_fourcc: FOURCC('G','R','E','Y'), 
      src_guid: { 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 
		  0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00 }, 
      src_bpp: 8, 
      
      dest_fourcc: FOURCC('I','4','2','0'), 
      dest_guid: GUID_UYVY, 
      dest_bpp: 12,
      
      priority: 5, 

      func: y8002y420,

      flags: 0
   },
   { 
      src_fourcc: FOURCC('I','4','2','0'), 
      src_guid: { 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 
		  0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00 }, 
      src_bpp: 12, 
      
      dest_fourcc: FOURCC('Y','8','0','0'), 
      dest_guid: { 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 
		  0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00 }, 
      dest_bpp: 8,
      
      priority: 5, 

      func: y4202y800,

      flags: 0
   },
   { 
      src_fourcc: FOURCC('Y','8','0','0'), 
      src_guid: { 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 
		  0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00 }, 
      src_bpp: 8, 
      
      dest_fourcc: FOURCC('I','4','2','2'), 
      dest_guid: GUID_UYVY, 
      dest_bpp: 16,
      
      priority: 5, 

      func: y8002y422,

      flags: 0
   },
   { 
      src_fourcc: FOURCC('G','R','E','Y'), 
      src_guid: { 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 
		  0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00 }, 
      src_bpp: 8, 
      
      dest_fourcc: FOURCC('I','4','2','2'), 
      dest_guid: GUID_UYVY, 
      dest_bpp: 16,
      
      priority: 5, 

      func: y8002y422,

      flags: 0
   },
   { 
      src_fourcc: FOURCC('G','R','E','Y'), 
      src_guid: { 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 
		  0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00 }, 
      src_bpp: 8, 
      
      dest_fourcc: FOURCC('Y','U','Y','2'), 
      dest_guid: GUID_UYVY, 
      dest_bpp: 16,
      
      priority: 5, 

      func: grey2yuy2,

      flags: 0
   },
   { 
      src_fourcc: FOURCC('Y','8','0','0'), 
      src_guid: { 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 
		  0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00 }, 
      src_bpp: 8, 
      
      dest_fourcc: FOURCC('Y','U','Y','2'), 
      dest_guid: GUID_UYVY, 
      dest_bpp: 16,
      
      priority: 5, 

      func: grey2yuy2,

      flags: 0
   },
   {
      src_fourcc: FOURCC('U','Y','V','Y'),
      src_guid: GUID_UYVY,
      src_bpp: 16, 
      
      dest_fourcc: FOURCC('I','4','2','2'),
      dest_guid: GUID_I422,
      dest_bpp: 16,

      priority: 15,
      func: uyvytoyuv422p,
      flags: 0
   },

   {
      src_fourcc: FOURCC('U','Y','V','Y'),
      src_guid: GUID_UYVY,
      src_bpp: 16, 
      
      dest_fourcc: FOURCC('I','4','2','0'),
      dest_guid: GUID_I420,
      dest_bpp: 12,

      priority: 15,
      func: uyvytoyuv420p,
      flags: 0
   },
   
   {
      src_fourcc: FOURCC('I','4','2','0'),
      src_guid: GUID_I420,
      src_bpp: 12,

      dest_fourcc: FOURCC('U','Y','V','Y'),
      dest_guid: GUID_UYVY,
      dest_bpp: 16, 
      
      priority: 15,
      func: yuv420ptouyvy,
      flags: 0      
   },

   {
      src_fourcc: FOURCC('Y','U','Y','V'),
      src_guid: GUID_UYVY,
      src_bpp: 16, 
      
      dest_fourcc: FOURCC('I','4','2','0'),
      dest_guid: GUID_I420,
      dest_bpp: 12,

      priority: 15,
      func: yuyvtoyuv420p,
      flags: 0
   },

   {
      src_fourcc: FOURCC('Y','U','Y','2'),
      src_guid: GUID_UYVY,
      src_bpp: 16, 
      
      dest_fourcc: FOURCC('I','4','2','0'),
      dest_guid: GUID_I420,
      dest_bpp: 12,

      priority: 15,
      func: yuyvtoyuv420p,
      flags: 0
   },
   
   {
      src_fourcc: FOURCC('I','4','2','0'),
      src_guid: GUID_I420,
      src_bpp: 12,

      dest_fourcc: FOURCC('Y','U','Y','V'),
      dest_guid: GUID_UYVY,
      dest_bpp: 16, 
      
      priority: 15,
      func: yuv420ptoyuyv,
      flags: 0      
   },

   {
      src_fourcc: FOURCC('I','4','2','0'),
      src_guid: GUID_I420,
      src_bpp: 12,

      dest_fourcc: FOURCC('R','G','B','3'),
      dest_guid: GUID_UYVY,
      dest_bpp: 24, 
      
      priority: 15,
      func: y4202rgb24,
      flags: 0      
   },

   {
      src_fourcc: FOURCC('I','4','2','0'),
      src_guid: GUID_I420,
      src_bpp: 12,

      dest_fourcc: FOURCC('R','G','B',0),
      dest_guid: GUID_RGB24,
      dest_bpp: 24, 
      
      priority: 15,
      func: y4202rgb24,
      flags: 0      
   },

   {
      src_fourcc: FOURCC('I','4','2','0'),
      src_guid: GUID_I420,
      src_bpp: 12,

      dest_fourcc: FOURCC('R','G','B','4'),
      dest_guid: GUID_RGB32,
      dest_bpp: 32, 
      
      priority: 15,
      func: y4202rgb32,
      flags: 0      
   },

   { 
      src_fourcc: FOURCC('Y','U','Y','V'), 
      src_guid: GUID_UYVY, 
      src_bpp: 16, 
      
      dest_fourcc: FOURCC('Y','U','Y','2'), 
      dest_guid: GUID_YUY2, 
      dest_bpp: 16,
      
      priority: 20, 

      func: NULL,

      flags: XFM_PASSTHROUGH
   },
   { 
      src_fourcc: FOURCC('Y','U','Y','V'), 
      src_guid: GUID_UYVY, 
      src_bpp: 16, 
      
      dest_fourcc: FOURCC('U','Y','V','Y'), 
      dest_guid: GUID_YUY2, 
      dest_bpp: 16,
      
      priority: 10, 

      func: uyvy2yuy2,

      flags: XFM_IN_PLACE
   },
   { 
      src_fourcc: FOURCC('U','Y','V','Y'), 
      src_guid: GUID_UYVY, 
      src_bpp: 16, 
      
      dest_fourcc: FOURCC('Y','U','Y','2'), 
      dest_guid: GUID_YUY2, 
      dest_bpp: 16,
      
      priority: 10, 

      func: uyvy2yuy2,

      flags: XFM_IN_PLACE
   },
   { 
      src_fourcc: FOURCC('Y','U','Y','2'), 
      src_guid: GUID_YUY2, 
      src_bpp: 16, 
      
      dest_fourcc: FOURCC('Y','U','Y','V'), 
      dest_guid: GUID_YUY2, 
      dest_bpp: 16,
      
      priority: 10, 

      func: yuy2toyuyv,

      flags: 0
   },
   { 
      src_fourcc: FOURCC('U','Y','V','Y'), 
      src_guid: GUID_UYVY, 
      src_bpp: 16, 
      
      dest_fourcc: FOURCC('Y','U','V',0), 
      dest_guid: GUID_YUV, 
      dest_bpp: 24,
      
      priority: 10, 

      func: uyvy2yuv,

      flags: 0,
   },
   // uyvy -> rgb24
   { 
      src_fourcc: FOURCC('U','Y','V','Y'), 
      src_guid: GUID_UYVY, 
      src_bpp: 16, 
      
      dest_fourcc: FOURCC('R','G','B',0), 
      dest_guid: GUID_RGB24, 
      dest_bpp: 24,
      
      priority: 5, 

      func: uyvy2rgb24,

      flags: 0
   },
   // uyvy -> rgb24
   { 
      src_fourcc: FOURCC('U','Y','V','Y'), 
      src_guid: GUID_UYVY, 
      src_bpp: 16, 
      
      dest_fourcc: FOURCC('R','G','B','3'), 
      dest_guid: GUID_RGB24, 
      dest_bpp: 24,
      
      priority: 5, 

      func: uyvy2rgb24,

      flags: 0
   },
   // yuyv -> rgb24
   { 
      src_fourcc: FOURCC('Y','U','Y','V'), 
      src_guid: GUID_YUY2, 
      src_bpp: 16, 
      
      dest_fourcc: FOURCC('R','G','B',0), 
      dest_guid: GUID_RGB24, 
      dest_bpp: 24,
      
      priority: 5, 

      func: yuyv2rgb24,

      flags: 0
   },
   // yuyv -> rgb24
   { 
      src_fourcc: FOURCC('Y','U','Y','2'), 
      src_guid: GUID_YUY2, 
      src_bpp: 16, 
      
      dest_fourcc: FOURCC('R','G','B',0), 
      dest_guid: GUID_RGB24, 
      dest_bpp: 24,
      
      priority: 5, 

      func: yuyv2rgb24,

      flags: 0
   },
   // yuyv -> rgb24
   { 
      src_fourcc: FOURCC('Y','U','Y','2'), 
      src_guid: GUID_YUY2, 
      src_bpp: 16, 
      
      dest_fourcc: FOURCC('R','G','B','3'), 
      dest_guid: GUID_RGB24, 
      dest_bpp: 24,
      
      priority: 5, 

      func: yuyv2rgb24,

      flags: 0
   },
   // uyvy -> rgb24
   { 
      src_fourcc: FOURCC('Y','U','Y','V'), 
      src_guid: GUID_UYVY, 
      src_bpp: 16, 
      
      dest_fourcc: FOURCC('R','G','B','3'), 
      dest_guid: GUID_RGB24, 
      dest_bpp: 24,
      
      priority: 5, 

      func: yuyv2rgb24,

      flags: 0
   },
   // yuyv -> y800
   { 
      src_fourcc: FOURCC('Y','U','Y','V'), 
      src_guid: GUID_UYVY, 
      src_bpp: 16, 
      
      dest_fourcc: FOURCC('Y','8','0','0'), 
      dest_guid: GUID_Y800, 
      dest_bpp: 8,
      
      priority: 5, 

      func: yuy22grey,

      flags: 0
   },
   // uyvy -> y800
   { 
      src_fourcc: FOURCC('U','Y','V','Y'), 
      src_guid: GUID_UYVY, 
      src_bpp: 16, 
      
      dest_fourcc: FOURCC('Y','8','0','0'), 
      dest_guid: GUID_Y800, 
      dest_bpp: 8,
      
      priority: 5, 

      func: uyvy2grey,

      flags: 0
   },


   { 
      src_fourcc: FOURCC('U','Y','V','Y'), 
      src_guid: GUID_UYVY, 
      src_bpp: 16, 
      
      dest_fourcc: FOURCC('R','G','B','A'), 
      dest_guid: GUID_RGB32, 
      dest_bpp: 32,
      
      priority: 5, 

      func: uyvy2rgb32,

      flags: 0
   },      
   { 
      src_fourcc: FOURCC('U','Y','V','Y'), 
      src_guid: GUID_UYVY, 
      src_bpp: 16, 
      
      dest_fourcc: FOURCC('R','G','B','4'), 
      dest_guid: GUID_RGB32, 
      dest_bpp: 32,
      
      priority: 5, 

      func: uyvy2rgb32,

      flags: 0
   },      
   {
      src_fourcc: FOURCC('Y','4','1','1'),
      src_guid: GUID_UYVY,
      src_bpp: 12,
      
      dest_fourcc: FOURCC('R','G','B',0),
      dest_guid: GUID_RGB24,
      dest_bpp: 24,
      
      priority: 5,

      func: y4112rgb24,

      flags: 0
   },
   { 
      src_fourcc: FOURCC('Y','4','1','1'), 
      src_guid: GUID_UYVY, 
      src_bpp: 12, 
      
      dest_fourcc: FOURCC('R','G','B','A'), 
      dest_guid: GUID_RGB32, 
      dest_bpp: 32,
      
      priority: 5, 

      func: y4112rgb32,

      flags: 0
   },      
   {
      src_fourcc: FOURCC('Y','8','0','0'),
      src_guid: GUID_UYVY,
      src_bpp: 8,
      
      dest_fourcc: FOURCC('R','G','B',0),
      dest_guid: GUID_RGB24,
      dest_bpp: 24,
      
      priority: 5,

      func: y8002rgb24,

      flags: 0
   },
   {
      src_fourcc: FOURCC('Y','8','0','0'),
      src_guid: GUID_UYVY,
      src_bpp: 8,
      
      dest_fourcc: FOURCC('R','G','B','3'),
      dest_guid: GUID_RGB24,
      dest_bpp: 24,
      
      priority: 5,

      func: y8002rgb24,

      flags: 0
   },

   {
      src_fourcc: FOURCC('G','R','E','Y'),
      src_guid: GUID_UYVY,
      src_bpp: 8,
      
      dest_fourcc: FOURCC('R','G','B',0),
      dest_guid: GUID_RGB24,
      dest_bpp: 24,
      
      priority: 5,

      func: y8002rgb24,

      flags: 0
   },
   {
      src_fourcc: FOURCC('G','R','E','Y'),
      src_guid: GUID_UYVY,
      src_bpp: 8,
      
      dest_fourcc: FOURCC('R','G','B','3'),
      dest_guid: GUID_RGB24,
      dest_bpp: 24,
      
      priority: 5,

      func: y8002rgb24,

      flags: 0
   },
   // Bt848 RAW, treat as Mono8
   {
      src_fourcc: FOURCC('R','A','W',' '),
      src_guid: GUID_UYVY,
      src_bpp: 8,
      
      dest_fourcc: FOURCC('R','G','B',0),
      dest_guid: GUID_RGB24,
      dest_bpp: 24,
      
      priority: 5,

      func: y8002rgb24,

      flags: 0
   },
   {
      src_fourcc: FOURCC('Y','8','0','0'),
      src_guid: GUID_UYVY,
      src_bpp: 8,
      
      dest_fourcc: FOURCC('R','G','B','A'),
      dest_guid: GUID_RGB32,
      dest_bpp: 32,
      
      priority: 5,

      func: y8002rgb32,

      flags: 0
   },
   { 
      src_fourcc: FOURCC('R','G','B','4'), 
      src_guid: GUID_RGB32, 
      src_bpp: 32, 
      
      dest_fourcc: FOURCC('R','G','B','A'), 
      dest_guid: GUID_RGB32, 
      dest_bpp: 32,
      
      priority: 20, 

      func: NULL,

      flags: XFM_PASSTHROUGH
   },
   // bgr24 -> uyvy
   { 
      src_fourcc: FOURCC('B','G','R','3'), 
      src_guid: GUID_RGB24, 
      src_bpp: 24,
      
      dest_fourcc: FOURCC('U','Y','V','Y'), 
      dest_guid: GUID_UYVY, 
      dest_bpp: 16, 
      
      priority: 5, 

      func: bgr242uyvy,

      flags: 0
   },
   // bgr24 -> uyvy
   { 
      src_fourcc: FOURCC('B','G','R',0), 
      src_guid: GUID_RGB24, 
      src_bpp: 24,
      
      dest_fourcc: FOURCC('U','Y','V','Y'), 
      dest_guid: GUID_UYVY, 
      dest_bpp: 16, 
      
      priority: 5, 

      func: bgr242uyvy,

      flags: 0
   },
   // rgb24 -> uyvy
   { 
      src_fourcc: FOURCC('R','G','B','3'), 
      src_guid: GUID_RGB24, 
      src_bpp: 24,
      
      dest_fourcc: FOURCC('U','Y','V','Y'), 
      dest_guid: GUID_UYVY, 
      dest_bpp: 16, 
      
      priority: 5, 

      func: rgb242uyvy,

      flags: 0
   },
   // rgb24 -> uyvy
   { 
      src_fourcc: FOURCC('R','G','B',0), 
      src_guid: GUID_RGB24, 
      src_bpp: 24,
      
      dest_fourcc: FOURCC('U','Y','V','Y'), 
      dest_guid: GUID_UYVY, 
      dest_bpp: 16, 
      
      priority: 5, 

      func: rgb242uyvy,

      flags: 0
   },      
   // rgb32 -> uyvy
   { 
      src_fourcc: FOURCC('R','G','B','A'), 
      src_guid: GUID_RGB32, 
      src_bpp: 32,
      
      dest_fourcc: FOURCC('U','Y','V','Y'), 
      dest_guid: GUID_UYVY, 
      dest_bpp: 16, 
      
      priority: 5, 

      func: rgb322uyvy,

      flags: 0
   },      
   // rgb32 -> uyvy
   { 
      src_fourcc: FOURCC('R','G','B','4'), 
      src_guid: GUID_RGB32, 
      src_bpp: 32,
      
      dest_fourcc: FOURCC('U','Y','V','Y'), 
      dest_guid: GUID_UYVY, 
      dest_bpp: 16, 
      
      priority: 5, 

      func: rgb322uyvy,

      flags: 0
   },      
   // rgb24 -> uyvy
   { 
      src_fourcc: FOURCC('R','G','B','3'), 
      src_guid: GUID_RGB24, 
      src_bpp: 24,
      
      dest_fourcc: FOURCC('I','4','2','0'), 
      dest_guid: GUID_UYVY, 
      dest_bpp: 12, 
      
      priority: 5, 

      func: rgb24toyuv420p,

      flags: 0
   },
   // rgb24 -> uyvy
   { 
      src_fourcc: FOURCC('R','G','B',0), 
      src_guid: GUID_RGB24, 
      src_bpp: 24,
      
      dest_fourcc: FOURCC('I','4','2','0'), 
      dest_guid: GUID_UYVY, 
      dest_bpp: 12, 
      
      priority: 5, 

      func: rgb24toyuv420p,

      flags: 0
   },      
   { 
      src_fourcc: FOURCC('B','G','R','3'), 
      src_guid: GUID_RGB24, 
      src_bpp: 24,
      
      dest_fourcc: FOURCC('R','G','B','3'), 
      dest_guid: GUID_RGB24, 
      dest_bpp: 24, 
      
      priority: 5, 

      func: bgr24torgb24,

      flags: 0
   },
   { 
      src_fourcc: FOURCC('R','G','B','4'), 
      src_guid: GUID_RGB32, 
      src_bpp: 32,
      
      dest_fourcc: FOURCC('R','G','B','3'), 
      dest_guid: GUID_RGB24, 
      dest_bpp: 24, 
      
      priority: 5, 

      func: rgb32torgb24,

      flags: 0
   },
   { 
      src_fourcc: FOURCC('B','Y','8',' '), 
      src_guid: 0, 
      src_bpp: 8,
      
      dest_fourcc: FOURCC('U','Y','V','Y'), 
      dest_guid: GUID_UYVY, 
      dest_bpp: 16, 
      
      priority: 5, 

      func: by8touyvy,

      flags: 0
   },
   { 
      src_fourcc: FOURCC('B','Y','8',' '), 
      src_guid: 0, 
      src_bpp: 8,
      
      dest_fourcc: FOURCC('R','G','B','3'), 
      dest_guid: GUID_RGB24, 
      dest_bpp: 24, 
      
      priority: 5, 

      func: by8torgb24,

      flags: 0
   },
   { 
      src_fourcc: FOURCC('B','Y','8',' '), 
      src_guid: 0, 
      src_bpp: 8,
      
      dest_fourcc: FOURCC('B','G','R','3'), 
      dest_guid: GUID_BGR24, 
      dest_bpp: 24, 
      
      priority: 5, 

      func: by8tobgr24,

      flags: 0
   },
   { 
      src_fourcc: FOURCC('B','Y','8',' '), 
      src_guid: 0, 
      src_bpp: 8,
      
      dest_fourcc: FOURCC('R','G','B','A'), 
      dest_guid: GUID_RGB32, 
      dest_bpp: 32, 
      
      priority: 5, 

      func: by8torgb32,

      flags: 0
   },
   { 
      src_fourcc: FOURCC('B','Y','8',' '), 
      src_guid: 0, 
      src_bpp: 8,
      
      dest_fourcc: FOURCC('U','Y','V','Y'), 
      dest_guid: GUID_UYVY, 
      dest_bpp: 16, 
      
      priority: 5, 

      func: by8touyvy,

      flags: 0
   },
};

#if HAVE_AVCODEC


/* for compatibility with older versions of libavcodec */

#ifndef PIX_FMT_YUYV422
#define PIX_FMT_YUYV422 PIX_FMT_YUV422
#endif

#ifndef PIX_FMT_RGB32
#define PIX_FMT_RGB32 PIX_FMT_RGBA32
#endif


int PixelFormatBPP[] = {
   //PIX_FMT_NONE= -1,
   12,// PIX_FMT_YUV420P,   ///< Planar YUV 4:2:0, 12bpp, (1 Cr & Cb sample per 2x2 Y samples)
   16,// PIX_FMT_YUYV422,   ///< Packed YUV 4:2:2, 16bpp, Y0 Cb Y1 Cr
   24,// PIX_FMT_RGB24,     ///< Packed RGB 8:8:8, 24bpp, RGBRGB...
   24,// PIX_FMT_BGR24,     ///< Packed RGB 8:8:8, 24bpp, BGRBGR...
   16,// PIX_FMT_YUV422P,   ///< Planar YUV 4:2:2, 16bpp, (1 Cr & Cb sample per 2x1 Y samples)
   24,// PIX_FMT_YUV444P,   ///< Planar YUV 4:4:4, 24bpp, (1 Cr & Cb sample per 1x1 Y samples)
   32,// PIX_FMT_RGB32,     ///< Packed RGB 8:8:8, 32bpp, (msb)8A 8R 8G 8B(lsb), in cpu endianness
   9,// PIX_FMT_YUV410P,   ///< Planar YUV 4:1:0,  9bpp, (1 Cr & Cb sample per 4x4 Y samples)
   12,// PIX_FMT_YUV411P,   ///< Planar YUV 4:1:1, 12bpp, (1 Cr & Cb sample per 4x1 Y samples)
   16,// PIX_FMT_RGB565,    ///< Packed RGB 5:6:5, 16bpp, (msb)   5R 6G 5B(lsb), in cpu endianness
   16,// PIX_FMT_RGB555,    ///< Packed RGB 5:5:5, 16bpp, (msb)1A 5R 5G 5B(lsb), in cpu endianness most significant bit to 1
   8,// PIX_FMT_GRAY8,     ///<        Y        ,  8bpp
   1,// PIX_FMT_MONOWHITE, ///<        Y        ,  1bpp, 1 is white
   1,// PIX_FMT_MONOBLACK, ///<        Y        ,  1bpp, 0 is black
   8,// PIX_FMT_PAL8,      ///< 8 bit with PIX_FMT_RGB32 palette
   12,// PIX_FMT_YUVJ420P,  ///< Planar YUV 4:2:0, 12bpp, full scale (jpeg)
   16,// PIX_FMT_YUVJ422P,  ///< Planar YUV 4:2:2, 16bpp, full scale (jpeg)
   24,// PIX_FMT_YUVJ444P,  ///< Planar YUV 4:4:4, 24bpp, full scale (jpeg)
   8,// PIX_FMT_XVMC_MPEG2_MC,///< XVideo Motion Acceleration via common packet passing(xvmc_render.h)
   8,// PIX_FMT_XVMC_MPEG2_IDCT,
   16,// PIX_FMT_UYVY422,   ///< Packed YUV 4:2:2, 16bpp, Cb Y0 Cr Y1
   12,// PIX_FMT_UYYVYY411, ///< Packed YUV 4:1:1, 12bpp, Cb Y0 Y1 Cr Y2 Y3
   // PIX_FMT_NB,        ///< number of pixel formats, DO NOT USE THIS if you want to link with shared libav* because the number of formats might differ between versions
};

typedef struct PixelFormatFourccMap
{
      unsigned int fourcc;
      int pix_fmt;
} PixelFormatFourccMap;

PixelFormatFourccMap PixelFormatFourccs[] = 
{
   { UCIL_FOURCC( 'Y','U','Y','V' ), PIX_FMT_YUYV422 },
   { UCIL_FOURCC( 'R','G','B','3' ), PIX_FMT_RGB24 },
   { UCIL_FOURCC( 'B','G','R','3' ), PIX_FMT_BGR24 },
   { UCIL_FOURCC( 'R','G','B','4' ), PIX_FMT_RGB32 },
   { UCIL_FOURCC( 'U','Y','V','Y' ), PIX_FMT_UYVY422 },
   { UCIL_FOURCC( 'Y','8','0','0' ), PIX_FMT_GRAY8 },
};

const int ucil_NumPixelFormatFourccs = sizeof( PixelFormatFourccs ) / sizeof( PixelFormatFourccMap );

static int get_av_pix_fmt_from_fourcc( unsigned int fourcc )
{
   int i;
   

   for( i = 0; i < ucil_NumPixelFormatFourccs; i++ )
   {
      if( PixelFormatFourccs[i].fourcc == fourcc )
      {
	 return PixelFormatFourccs[i].pix_fmt;
      }
   }

   return -1;
}

#endif //HAVE_AVCODEC


static void ucil_get_xfminfo_from_fourcc( unsigned int src_fourcc, 
					  unsigned int dest_fourcc, 
					  xfm_info_t *info )
{
   int nconvs = sizeof( conversions ) / sizeof( xfm_info_t );
   int priority = -1;
   int i, found = -1;

   for( i = 0; i < nconvs; i++ )
   {
      if( ( conversions[i].src_fourcc == src_fourcc ) && 
	  ( conversions[i].dest_fourcc == dest_fourcc ) && 
	  ( conversions[i].priority > priority ) )
      {
	 priority = conversions[i].priority;
	 found = i;
      }
   }
   
   if( found == -1 )
   {
      info->src_fourcc = src_fourcc;
      memset( info->src_guid, 0x0, sizeof( info->src_guid ) );
      info->dest_fourcc = dest_fourcc;
      memset( info->dest_guid, 0x0, sizeof( info->dest_guid ) );
      info->func = NULL;
      info->flags = XFM_PASSTHROUGH;

      if( src_fourcc == dest_fourcc )
      {
	 info->priority = 20;
      }
      else
      {
	 info->priority = 1;
	 info->flags = XFM_NO_CONVERSION;
      }
   }
   else
   {
      memcpy( info, &conversions[found], sizeof( xfm_info_t ) );
   }
}

unicap_status_t ucil_convert_buffer( unicap_data_buffer_t *dest, unicap_data_buffer_t *src )
{
   unicap_status_t status = STATUS_FAILURE;

   if( src->format.fourcc == dest->format.fourcc )
   {
      memcpy( dest->data, src->data, src->format.buffer_size );
      status = STATUS_SUCCESS;
   }
   else
   {
      if( ( xfminfo.src_fourcc != src->format.fourcc ) ||
	  ( xfminfo.dest_fourcc != src->format.fourcc ) || 
	  ( xfminfo.flags & XFM_NO_CONVERSION ) )
      {
	 ucil_get_xfminfo_from_fourcc( src->format.fourcc, dest->format.fourcc, &xfminfo );      
      }
      
      if( !(xfminfo.flags & XFM_NO_CONVERSION) )
      {
	 dest->format.bpp = xfminfo.dest_bpp;
	 if( xfminfo.flags & XFM_PASSTHROUGH )
	 {
	    memcpy( dest->data, src->data, src->format.buffer_size );
	    status = STATUS_SUCCESS;
	 }
	 else
	 {
	    xfminfo.func( dest->data, src->data, src->format.size.width, src->format.size.height );
	    status = STATUS_SUCCESS;
	 }

      }
      else
      {
	 g_message( "Could not convert format: %s %08x to %s %08x\n", 
		    src->format.identifier, src->format.fourcc,
		    dest->format.identifier, dest->format.fourcc );
      }
      
   }
}

int ucil_conversion_supported( unsigned int dest_fourcc, unsigned int src_fourcc )
{
   int ret = 0;
   
   if( dest_fourcc == src_fourcc )
   {
      ret = 1;
   }
   else
   {
      xfm_info_t tmpinfo;
      
      ucil_get_xfminfo_from_fourcc( src_fourcc, dest_fourcc, &tmpinfo );
      if( !(tmpinfo.flags & XFM_NO_CONVERSION ) )
      {
	 ret = 1;
      }
   }
   
   return ret;
}


static void uyvytoyuv420p( __u8 *dest, __u8 *src, int width, int height )
{
   int size;
   int offy = 0;
   int offu = 0;
   int offv = 0;
   int offsrc = 0;
   int x,y;
   int stride;

   __u8 *dest_u;
   __u8 *dest_v;
   __u8 *dest_y;

   size = width * height;
   stride = width * 2;
   dest_y = dest;
   dest_u = dest + size;
   dest_v = dest + size + size / 4;
   
   for( y = 0; y < height; y += 2 )
   {
      int hoff = y * stride;
      for( x = 0; x < stride; x+=2 )
      {
	 dest_y[offy++] = src[hoff + 1 + x ];
      }
      for( x = 0; x < stride; x+=2 )
      {
	 dest_y[offy++] = src[hoff + 1 + x + stride ];
      }
      for( x = 0; x < stride; x+=4 )
      {
	 int tmpoffset = hoff + x;
	 dest_u[offu++] = src[ tmpoffset ];
	 dest_v[offv++] = src[ tmpoffset + 2 ];
      }
   }
}

static void yuv420ptouyvy( __u8 *dest, __u8 *src, int width, int height )
{
   int ysize = width*height;
   int offy = 0;
   int offu = 0;
   int offv = 0;
   int offdest = 0;
   int y;
   int stride;

   stride = width * 2;

   __u8 *src_y = src;
   __u8 *src_u = src + ysize;
   __u8 *src_v = src + ysize + ysize / 4;

   for( y = 0; y < height; y += 2 )
   {
      int hoff = stride * y;
      int x;
      
      for( x = 0; x < stride; x+=4 )
      {
	 __u8 u = src_u[ offu++ ];
	 __u8 v = src_v[ offv++ ];

	 dest[ hoff + x ] = u;
	 dest[ hoff + stride + x ] = u;

	 dest[ hoff + x + 2 ] = v;
	 dest[ hoff + stride + x + 2 ] = v;
	 
      }
      

      for( x = 0; x < stride; x += 2 )
      {
	 dest[ hoff + x + 1 ] = src_y[ offy++ ];
      }

      for( x = 0; x < stride; x+= 2 )
      {
	 dest[ hoff + stride + x + 1 ] = src_y[ offy++ ];
      }
   }
}

static void uyvytoyuv422p( __u8 *dest, __u8 *src, int width, int height )
{
   int size;
   int offy = 0;
   int offu = 0;
   int offv = 0;
   int offsrc = 0;
   __u8 *dest_u;
   __u8 *dest_v;
   __u8 *dest_y;

   size = width * height;
   dest_y = dest;
   dest_u = dest + size;
   dest_v = dest + size + size / 2;
   
   while( offsrc < (size*2)  )
   {
      dest_u[offu++] = src[offsrc++];
      dest_y[offy++] = src[offsrc++];
      dest_v[offv++] = src[offsrc++];
      dest_y[offy++] = src[offsrc++];
   }
}


static void yuyvtoyuv420p( __u8 *dest, __u8 *src, int width, int height )
{
   int size;
   int offy = 0;
   int offu = 0;
   int offv = 0;
   int offsrc = 0;
   int x,y;
   int stride;

   __u8 *dest_u;
   __u8 *dest_v;
   __u8 *dest_y;

   size = width * height;
   stride = width * 2;
   dest_y = dest;
   dest_u = dest + size;
   dest_v = dest + size + size / 4;
   
   for( y = 0; y < height; y += 2 )
   {
      int hoff = y * stride;
      for( x = 0; x < stride; x+=2 )
      {
	 dest_y[offy++] = src[hoff + x ];
      }
      for( x = 0; x < stride; x+=2 )
      {
	 dest_y[offy++] = src[hoff + x + stride ];
      }
      for( x = 0; x < stride; x+=4 )
      {
	 int tmpoffset = hoff + x;
	 dest_u[offu++] = src[ tmpoffset + 1 ];
	 dest_v[offv++] = src[ tmpoffset + 3 ];
      }
   }
}

static void yuv420ptoyuyv( __u8 *dest, __u8 *src, int width, int height )
{
   int ysize = width*height;
   int offy = 0;
   int offu = 0;
   int offv = 0;
   int offdest = 0;
   int y;
   int stride;

   stride = width * 2;

   __u8 *src_y = src;
   __u8 *src_u = src + ysize;
   __u8 *src_v = src + ysize + ysize / 4;

   for( y = 0; y < height; y += 2 )
   {
      int hoff = stride * y;
      int x;
      
      for( x = 0; x < stride; x+=4 )
      {
	 __u8 u = src_u[ offu++ ];
	 __u8 v = src_v[ offv++ ];

	 dest[ hoff + x + 1 ] = u;
	 dest[ hoff + stride + x + 1 ] = u;

	 dest[ hoff + x + 3 ] = v;
	 dest[ hoff + stride + x + 3 ] = v;
	 
      }
      

      for( x = 0; x < stride; x += 2 )
      {
	 dest[ hoff + x ] = src_y[ offy++ ];
      }

      for( x = 0; x < stride; x+= 2 )
      {
	 dest[ hoff + stride + x ] = src_y[ offy++ ];
      }
   }
}

static void grey2uyvy( __u8 *dest, __u8 *source, int width, int height )
{
   int i,j;

   int source_size = width * height;

   for( i = 0, j = 0; i < source_size; i+=2, j+=4 )
   {
      __u32 tmp = 0x7f007f00 | ( source[i]<<16 ) | source[i+1];
      *(int*)(dest+j) = GINT32_FROM_BE( tmp );
   }
}

static void uyvy2grey( __u8 *dest, __u8 *source, int width, int height )
{
   int i,j;

   int source_size = width * height * 2;

   for( i = 1, j = 0; i < source_size; i += 2, j++ )
   {
      dest[j] = source[i];
   }
}

static void grey2yuy2( __u8 *dest, __u8 *source, int width, int height )
{
   int i,j;

   int source_size = width * height;


   for( i = 0, j = 0; i < source_size; i+=2, j+=4 )
   {
      __u32 tmp = 0x007f007f + 
	 ( ((unsigned int)source[i])<<24 ) +
	 (((unsigned int)source[i+1])<<8);
      *(int*)(dest+j) = GINT32_FROM_BE( tmp );
   }
}

static void yuy22grey( __u8 *dest, __u8 *source, int width, int height )
{
   int i,j;

   int source_size = width * height * 2;

   for( i = 0, j = 0; i < source_size; i += 2, j++ )
   {
      dest[j] = source[i];
   }
}
		
static void y8002y420( __u8 *dest, __u8 *source, int width, int height )
{
   memcpy( dest, source, width * height );
   memset( dest + width * height, 0x7f, width * height / 2 );
}

static void y4202y800( __u8 *dest, __u8 *source, int width, int height )
{
   memcpy( dest, source, width * height );
}

static void y8002y422( __u8 *dest, __u8 *source, int width, int height )
{
   memcpy( dest, source, width * height );
   memset( dest + width * height, 0x7f, width *  height );
}

static void uyvy2yuy2( __u8 *dest, __u8 *source, int width, int height )
{
   __u8 *end;

   int source_size = width * height * 2;

   for( end = source + source_size; source < end; source+=4, dest += 4 )
   {
      register __u32 tmp = *((int*)source);
      tmp = ( ( tmp & 0xff00ff00 ) >> 8 ) | ( ( tmp & 0xff00ff ) << 8 );
      *((int*)dest) = tmp;
   }   
}

static void uyvy2yuv( __u8 *dest, __u8 *source, int width, int height )
{
   __u8 *end = source + width * height * 2;

   while( source < end )
   {
      __u8 y1, y2, u, v;
      
      u = *source++;
      y1 = *source++;
      v = *source++;
      y2 = *source++;
      
      *dest++ = y1;
      *dest++ = u;
      *dest++ = v;
      *dest++ = y2;
      *dest++ = u;
      *dest++ = v;
   }
}

static void uyvy2rgb24( __u8 *dest, __u8 *source, int width, int height )
{
   __u8 *source_end = source + (width * height * 2);

   while( source < source_end )
   {
      __u8 y1, y2, u, v;
      int c1, c2, d, dg, db, e, er, eg;
      int ir, ig, ib;
		
      u = *source++;
      y1 = *source++;
      v = *source++;
      y2 = *source++;
      
      c1 = (y1-16)*298;
      c2 = (y2-16)*298;
      d = u-128;
      dg = 100 * d;
      db = 516 * d;
      e = v-128;
      er = 409 * e;
      eg = 208 * e;

      ir = (c1 + er + 128)>>8;
      ig = (c1 - dg - eg + 128 )>>8;
      ib = (c1 + db)>>8;
      

      *dest++ = (__u8) ( ir > 255 ? 255 : ( ir < 0 ? 0 : ir ) );
      *dest++ = (__u8) ( ig > 255 ? 255 : ( ig < 0 ? 0 : ig ) );
      *dest++ = (__u8) ( ib > 255 ? 255 : ( ib < 0 ? 0 : ib ) );
		
      ir = (c2 + er + 128)>>8;
      ig = (c2 - dg - eg + 128 )>>8;
      ib = (c2 + db)>>8;

      *dest++ = (__u8) ( ir > 255 ? 255 : ( ir < 0 ? 0 : ir ) );
      *dest++ = (__u8) ( ig > 255 ? 255 : ( ig < 0 ? 0 : ig ) );
      *dest++ = (__u8) ( ib > 255 ? 255 : ( ib < 0 ? 0 : ib ) );
   }
}

static void yuyv2rgb24( __u8 *dest, __u8 *source, int width, int height )
{
   __u8 *source_end = source + (width * height * 2);

   while( source < source_end )
   {
      __u8 y1, y2, u, v;
      int c1, c2, d, dg, db, e, er, eg;
      int ir, ig, ib;
		
      y1 = *source++;
      u = *source++;
      y2 = *source++;
      v = *source++;
      
      c1 = (y1-16)*298;
      c2 = (y2-16)*298;
      d = u-128;
      dg = 100 * d;
      db = 516 * d;
      e = v-128;
      er = 409 * e;
      eg = 208 * e;

      ir = (c1 + er + 128)>>8;
      ig = (c1 - dg - eg + 128 )>>8;
      ib = (c1 + db)>>8;
      

      *dest++ = (__u8) ( ir > 255 ? 255 : ( ir < 0 ? 0 : ir ) );
      *dest++ = (__u8) ( ig > 255 ? 255 : ( ig < 0 ? 0 : ig ) );
      *dest++ = (__u8) ( ib > 255 ? 255 : ( ib < 0 ? 0 : ib ) );
		
      ir = (c2 + er + 128)>>8;
      ig = (c2 - dg - eg + 128 )>>8;
      ib = (c2 + db)>>8;

      *dest++ = (__u8) ( ir > 255 ? 255 : ( ir < 0 ? 0 : ir ) );
      *dest++ = (__u8) ( ig > 255 ? 255 : ( ig < 0 ? 0 : ig ) );
      *dest++ = (__u8) ( ib > 255 ? 255 : ( ib < 0 ? 0 : ib ) );
   }
}


/* static void uyvy2rgb24( __u8 *dest, __u8 *source, int width, int height ) */
/* { */
/*    int i; */
/*    int dest_offset = 0; */
	
/*    int source_size = width * height * 2; */

/*    for( i = 0; i < source_size; i+=4 ) */
/*    { */
/*       __u8 *r, *b, *g; */
/*       __u8 *y1, *y2, *u, *v; */
		
/*       float fr, fg, fb; */
/*       float fy1, fy2, fu, fur, fug, fub, fv, fvr, fvg, fvb; */

/*       r = dest + dest_offset; */
/*       g = r + 1; */
/*       b = g + 1; */
		
/*       u = source + i; */
/*       y1 = u + 1; */
/*       v = y1 + 1; */
/*       y2 = v + 1; */
		
/*       fu = *u - 128; */
/*       fur = fu * 0.0009267; */
/*       fug = fu * 0.3436954; */
/*       fub = fu * 1.7721604; */
/*       fv = *v - 128; */
/*       fvr = fv * 1.4016868; */
/*       fvg = fv * 0.7141690; */
/*       fvb = fv * 0.0009902; */
/*       fy1= *y1; */
/*       fy2= *y2; */

/*       fr = fy1 - fur + fvr; */
/*       fg = fy1 - fug - fvg; */
/*       fb = fy1 + fub + fvb; */

/*       *r = (__u8) ( fr > 255 ? 255 : ( fr < 0 ? 0 : fr ) ); */
/*       *g = (__u8) ( fg > 255 ? 255 : ( fg < 0 ? 0 : fg ) ); */
/*       *b = (__u8) ( fb > 255 ? 255 : ( fb < 0 ? 0 : fb ) ); */


		
/*       dest_offset += 3; */
		
/*       r = dest + dest_offset; */
/*       g = r + 1; */
/*       b = g + 1; */
		
/*       fr = fy2 - fur + fvr; */
/*       fg = fy2 - fug - fvg; */
/*       fb = fy2 + fub + fvb; */

/*       *r = (__u8) ( fr > 255 ? 255 : ( fr < 0 ? 0 : fr ) ); */
/*       *g = (__u8) ( fg > 255 ? 255 : ( fg < 0 ? 0 : fg ) ); */
/*       *b = (__u8) ( fb > 255 ? 255 : ( fb < 0 ? 0 : fb ) ); */

/*       dest_offset += 3; */
/*    } */
/* } */


size_t __y4112rgb24( __u8 *dest, __u8 *source, size_t dest_size, size_t source_size )
{
   int i;
   int dest_offset = 0;
	
   long co1, co2, co3, co4, co5, co6;
	
   co1 = 0.0009267 * 65536;
   co2 = 1.4016868 * 65536;
   co3 = 0.3436954 * 65536;
   co4 = 0.7141690 * 65536;
   co5 = 1.7721604 * 65536;
   co6 = 0.0009902 * 65536;

   if( dest_size < ( source_size + ( source_size >> 1 ) ) )
   {
      return 0;
   }

   for( i = 0; i < source_size;  )
   {
      __u8 *r, *b, *g;
      long y1, y2, y3, y4, u, v;

      r = dest + dest_offset;
      g = r + 1;
      b = g + 1;
		
      u =  ( ( *( source + i++ ) ) << 16 ) - ( 128 << 16 );
      y1 = ( *( source + i++ ) ) << 16;
      y2 = ( *( source + i++ ) ) << 16;
      v =  ( ( *( source + i++ ) ) << 16 ) - ( 128 << 16 );
      y3 = ( *( source + i++ ) ) << 16;
      y4 = ( *( source + i++ ) ) << 16;

      *r = ( ( y1 - co1 * u + co2 * v ) >> 16 );
      *g = ( ( y1 - co3 * u - co4 * v ) >> 16 );
      *b = ( ( y1 + co5 * u + co6 * v ) >> 16 );
		
      dest_offset += 3;

      r = dest + dest_offset;
      g = r + 1;
      b = g + 1;
		
      *r = ( ( y2 - co1 * u + co2 * v ) >> 16 );
      *g = ( ( y2 - co3 * u - co4 * v ) >> 16 );
      *b = ( ( y2 + co5 * u + co6 * v ) >> 16 );

      dest_offset += 3;

      r = dest + dest_offset;
      g = r + 1;
      b = g + 1;
		
      *r = ( y3 - co1 * u + co2 * v ) >> 16;
      *g = ( y3 - co3 * u - co4 * v ) >> 16;
      *b = ( y3 + co5 * u + co6 * v ) >> 16;

      dest_offset += 3;

      r = dest + dest_offset;
      g = r + 1;
      b = g + 1;

      *r = ( y4 - co1 * u + co2 * v ) >> 16;
      *g = ( y4 - co3 * u - co4 * v ) >> 16;
      *b = ( y4 + co5 * u + co6 * v ) >> 16;

      dest_offset += 3;
   }
	
   return source_size * 2;
}


static void y4112rgb24( __u8 *dest, __u8 *source, int width, int height )
{
   int i;
   int dest_offset = 0;
	
   int source_size = width * height * 1.5;

   for( i = 0; i < source_size; i+=6 )
   {
      __u8 *r, *b, *g;
      __u8 *y1, *y2, *y3, *y4, *u, *v;
		
      double fr, fg, fb;
      double fy1, fy2, fy3, fy4, fu, fur, fug, fub, fv, fvr, fvg, fvb;

      r = dest + dest_offset;
      g = r + 1;
      b = g + 1;
		
      u = source + i;
      y1 = u + 1;
      y2 = y1 + 1;
      v = y2 + 1;
      y3 = v + 1;
      y4 = y3 + 1;

      fu = *u - 128;
      fur = fu * 0.0009267;
      fug = fu * 0.3436954;
      fub = fu * 1.7721604;
      fv = *v - 128;
      fvr = fv * 1.4016868;
      fvg = fv * 0.7141690;
      fvb = fv * 0.0009902;
      fy1= *y1;
      fy2= *y2;
      fy3= *y3;
      fy4= *y4;

      fr = fy1 - fur + fvr;
      fg = fy1 - fug - fvg;
      fb = fy1 + fub + fvb;

      *r = (__u8) ( fr > 255 ? 255 : ( fr < 0 ? 0 : fr ) );
      *g = (__u8) ( fg > 255 ? 255 : ( fg < 0 ? 0 : fg ) );
      *b = (__u8) ( fb > 255 ? 255 : ( fb < 0 ? 0 : fb ) );
		
		
      dest_offset += 3;
		
      r = dest + dest_offset;
      g = r + 1;
      b = g + 1;
		
      fr = fy2 - fur + fvr;
      fg = fy2 - fug - fvg;
      fb = fy2 + fub + fvb;

      *r = (__u8) ( fr > 255 ? 255 : ( fr < 0 ? 0 : fr ) );
      *g = (__u8) ( fg > 255 ? 255 : ( fg < 0 ? 0 : fg ) );
      *b = (__u8) ( fb > 255 ? 255 : ( fb < 0 ? 0 : fb ) );

      dest_offset += 3;

      r = dest + dest_offset;
      g = r + 1;
      b = g + 1;
		
      fr = fy3 - fur + fvr;
      fg = fy3 - fug - fvg;
      fb = fy3 + fub + fvb;

      *r = (__u8) ( fr > 255 ? 255 : ( fr < 0 ? 0 : fr ) );
      *g = (__u8) ( fg > 255 ? 255 : ( fg < 0 ? 0 : fg ) );
      *b = (__u8) ( fb > 255 ? 255 : ( fb < 0 ? 0 : fb ) );

      dest_offset += 3;

      r = dest + dest_offset;
      g = r + 1;
      b = g + 1;
		
      fr = fy4 - fur + fvr;
      fg = fy4 - fug - fvg;
      fb = fy4 + fub + fvb;

      *r = (__u8) ( fr > 255 ? 255 : ( fr < 0 ? 0 : fr ) );
      *g = (__u8) ( fg > 255 ? 255 : ( fg < 0 ? 0 : fg ) );
      *b = (__u8) ( fb > 255 ? 255 : ( fb < 0 ? 0 : fb ) );

      dest_offset += 3;
   }
}

static void y4112rgb32( __u8 *dest, __u8 *source, int width, int height )
{
   int i;
   int dest_offset = 0;
	
   int source_size = width * height * 1.5;

   for( i = 0; i < source_size; i+=6 )
   {
      __u8 *r, *b, *g, *a;
      __u8 *y1, *y2, *y3, *y4, *u, *v;
		
      double fr, fg, fb;
      double fy1, fy2, fy3, fy4, fu, fur, fug, fub, fv, fvr, fvg, fvb;

      r = dest + dest_offset;
      g = r + 1;
      b = g + 1;
      a = b + 1;
		
      u = source + i;
      y1 = u + 1;
      y2 = y1 + 1;
      v = y2 + 1;
      y3 = v + 1;
      y4 = y3 + 1;

      fu = *u - 128;
      fur = fu * 0.0009267;
      fug = fu * 0.3436954;
      fub = fu * 1.7721604;
      fv = *v - 128;
      fvr = fv * 1.4016868;
      fvg = fv * 0.7141690;
      fvb = fv * 0.0009902;
      fy1= *y1;
      fy2= *y2;
      fy3= *y3;
      fy4= *y4;

      fr = fy1 - fur + fvr;
      fg = fy1 - fug - fvg;
      fb = fy1 + fub + fvb;

      *r = (__u8) ( fr > 255 ? 255 : ( fr < 0 ? 0 : fr ) );
      *g = (__u8) ( fg > 255 ? 255 : ( fg < 0 ? 0 : fg ) );
      *b = (__u8) ( fb > 255 ? 255 : ( fb < 0 ? 0 : fb ) );
      *a = 0xff;
		
		
      dest_offset += 4;
		
      r = dest + dest_offset;
      g = r + 1;
      b = g + 1;
      a = b + 1;
		
      fr = fy2 - fur + fvr;
      fg = fy2 - fug - fvg;
      fb = fy2 + fub + fvb;

      *r = (__u8) ( fr > 255 ? 255 : ( fr < 0 ? 0 : fr ) );
      *g = (__u8) ( fg > 255 ? 255 : ( fg < 0 ? 0 : fg ) );
      *b = (__u8) ( fb > 255 ? 255 : ( fb < 0 ? 0 : fb ) );
      *a = 0xff;

      dest_offset += 4;

      r = dest + dest_offset;
      g = r + 1;
      b = g + 1;
      a = b + 1;
		
      fr = fy3 - fur + fvr;
      fg = fy3 - fug - fvg;
      fb = fy3 + fub + fvb;

      *r = (__u8) ( fr > 255 ? 255 : ( fr < 0 ? 0 : fr ) );
      *g = (__u8) ( fg > 255 ? 255 : ( fg < 0 ? 0 : fg ) );
      *b = (__u8) ( fb > 255 ? 255 : ( fb < 0 ? 0 : fb ) );
      *a = 0xff;

      dest_offset += 4;

      r = dest + dest_offset;
      g = r + 1;
      b = g + 1;
      a = b + 1;
		
      fr = fy4 - fur + fvr;
      fg = fy4 - fug - fvg;
      fb = fy4 + fub + fvb;

      *r = (__u8) ( fr > 255 ? 255 : ( fr < 0 ? 0 : fr ) );
      *g = (__u8) ( fg > 255 ? 255 : ( fg < 0 ? 0 : fg ) );
      *b = (__u8) ( fb > 255 ? 255 : ( fb < 0 ? 0 : fb ) );
      *a = 0xff;

      dest_offset += 4;
   }
}

static void uyvy2rgb32( __u8 *dest, __u8 *source, int width, int height )
{
   __u8 *source_end = source + (width * height * 2);

   while( source < source_end )
   {
      __u8 y1, y2, u, v;
      int c1, c2, d, dg, db, e, er, eg;
      int ir, ig, ib;		
		
      u = *source++;
      y1 = *source++;
      v = *source++;
      y2 = *source++;

      c1 = (y1-16)*298;
      c2 = (y2-16)*298;
      d = u-128;
      dg = 100 * d;
      db = 516 * d;
      e = v-128;
      er = 409 * e;
      eg = 208 * e;

      ir = (c1 + er + 128)>>8;
      ig = (c1 - dg - eg + 128 )>>8;
      ib = (c1 + db)>>8;
      

      *dest++ = (__u8) ( ir > 255 ? 255 : ( ir < 0 ? 0 : ir ) );
      *dest++ = (__u8) ( ig > 255 ? 255 : ( ig < 0 ? 0 : ig ) );
      *dest++ = (__u8) ( ib > 255 ? 255 : ( ib < 0 ? 0 : ib ) );
      *dest++ = 0;		

      ir = (c2 + er + 128)>>8;
      ig = (c2 - dg - eg + 128 )>>8;
      ib = (c2 + db)>>8;
      
      *dest++ = (__u8) ( ir > 255 ? 255 : ( ir < 0 ? 0 : ir ) );
      *dest++ = (__u8) ( ig > 255 ? 255 : ( ig < 0 ? 0 : ig ) );
      *dest++ = (__u8) ( ib > 255 ? 255 : ( ib < 0 ? 0 : ib ) );
      *dest++ = 0;		

   }
	
}

static void y4202rgb24( __u8 *dest, __u8 *source, int width, int height )
{
   int i, j;
   int dest_offset = 0;
   int line_offset;
   __u8 *ybase = source;
   __u8 *ubase = source + width * height;
   __u8 *vbase = ubase + ( ( width * height ) / 4 );
	
   for( j = 0; j < height; j++ )
   {
      for( i = 0; i < width; i+=2 )
      {
	 __u8 y, u, v;
	 int c, d, dg, db, e, er, eg;
	 int ir, ig, ib;

	 int uvidx = j /2 * width /2 + i /2;
			
	 y = *(ybase + ( j * width ) + i);
	 u = *(ubase + uvidx );
	 v = *(vbase + uvidx );
			
	 c = (y-16)*298;
	 d = u-128;
	 dg = 100 * d;
	 db = 516 * d;
	 e = v-128;
	 er = 409 * e;
	 eg = 208 * e;

	 ir = (c + er + 128)>>8;
	 ig = (c - dg - eg + 128 )>>8;
	 ib = (c + db)>>8;

	 *dest++ = (__u8) ( ir > 255 ? 255 : ( ir < 0 ? 0 : ir ) );
	 *dest++ = (__u8) ( ig > 255 ? 255 : ( ig < 0 ? 0 : ig ) );
	 *dest++ = (__u8) ( ib > 255 ? 255 : ( ib < 0 ? 0 : ib ) );

	 y = *(ybase + ( j * width ) + i + 1);

	 c = (y-16)*298;

	 ir = (c + er + 128)>>8;
	 ig = (c - dg - eg + 128 )>>8;
	 ib = (c + db)>>8;

	 *dest++ = (__u8) ( ir > 255 ? 255 : ( ir < 0 ? 0 : ir ) );
	 *dest++ = (__u8) ( ig > 255 ? 255 : ( ig < 0 ? 0 : ig ) );
	 *dest++ = (__u8) ( ib > 255 ? 255 : ( ib < 0 ? 0 : ib ) );

      }
   }
}

static void y4202rgb32( __u8 *dest, __u8 *source, int width, int height )
{
   int i, j;
   int dest_offset = 0;
   int line_offset;
   __u8 *ybase = source;
   __u8 *ubase = source + width * height;
   __u8 *vbase = ubase + ( ( width * height ) / 4 );
	
   for( j = 0; j < height; j++ )
   {
      for( i = 0; i < width; i+=2 )
      {
	 __u8 y, u, v;
	 int c, d, dg, db, e, er, eg;
	 int ir, ig, ib;

	 int uvidx = j /2 * width /2 + i /2;
			
	 y = *(ybase + ( j * width ) + i);
	 u = *(ubase + uvidx );
	 v = *(vbase + uvidx );
			
	 c = (y-16)*298;
	 d = u-128;
	 dg = 100 * d;
	 db = 516 * d;
	 e = v-128;
	 er = 409 * e;
	 eg = 208 * e;

	 ir = (c + er + 128)>>8;
	 ig = (c - dg - eg + 128 )>>8;
	 ib = (c + db)>>8;

	 *dest++ = (__u8) ( ir > 255 ? 255 : ( ir < 0 ? 0 : ir ) );
	 *dest++ = (__u8) ( ig > 255 ? 255 : ( ig < 0 ? 0 : ig ) );
	 *dest++ = (__u8) ( ib > 255 ? 255 : ( ib < 0 ? 0 : ib ) );
	 *dest++ = 0;// a

	 y = *(ybase + ( j * width ) + i + 1);

	 c = (y-16)*298;

	 ir = (c + er + 128)>>8;
	 ig = (c - dg - eg + 128 )>>8;
	 ib = (c + db)>>8;

	 *dest++ = (__u8) ( ir > 255 ? 255 : ( ir < 0 ? 0 : ir ) );
	 *dest++ = (__u8) ( ig > 255 ? 255 : ( ig < 0 ? 0 : ig ) );
	 *dest++ = (__u8) ( ib > 255 ? 255 : ( ib < 0 ? 0 : ib ) );
	 *dest++ = 0;// a

      }
   }
}

static void y8002rgb24( __u8 *dest, __u8 *source, int width, int height )
{
   int i;
   int dest_offset = 0;	
   int source_size = width * height;
	
   for( i = 0; i < source_size; i++ )
   {
      __u8 *y;
      __u8 *r, *b, *g;
		
      r = dest + dest_offset;
      g = r + 1;
      b = g + 1;
		
      y = source + i;
		
      *r = *y;
      *g = *y;
      *b = *y;
		
      dest_offset +=3;
   }
	
}

static void y8002rgb32( __u8 *dest, __u8 *source, int width, int height )
{
   int i;
   int dest_offset = 0;
   int source_size = width * height;
	
   for( i = 0; i < source_size; i++ )
   {
      __u8 *y;
      __u8 *r, *b, *g, *a;
		
      r = dest + dest_offset;
      g = r + 1;
      b = g + 1;
      a = b + 1;
		
      y = source + i;
		
      *r = *y;
      *g = *y;
      *b = *y;
      *a = 0;
		
      dest_offset +=4;
   }
}

//y =  0.2990 * R + 0.5870 * G + 0.1140 * B
//U = -0.1687 * R - 0.3313 * G + 0.5000 * B
//V =  0.5000 * R - 0.4187 * G - 0.0813 * B	
static void rgb242uyvy( __u8 *dest, __u8 *source, int width, int height )
{
   int offset;
   int dest_offset = 0;
   int source_size = width * height * 3;
   
   for( offset = 0; offset < source_size; )
   {
      int r, b, g;
      __u8 y;
      
      r = source[offset++];
      g = source[offset++];
      b = source[offset++];
      
      y = 0.299 * r + 0.587 * g + 0.114 * b;
      dest[dest_offset++] = 0.492 * (float)(b-y);
      dest[dest_offset++] = y;

      r = source[offset++];
      g = source[offset++];
      b = source[offset++];
      
      y = 0.299 * r + 0.587 * g + 0.114 * b;
      dest[dest_offset++] = 0.877 * (float)(r-y);
      dest[dest_offset++] = y;
   }
}

static void bgr242uyvy( __u8 *dest, __u8 *source, int width, int height )
{
   int offset;
   int dest_offset = 0;
   int source_size = width * height * 3;
   
   for( offset = 0; offset < source_size; )
   {
      int r, b, g;
      __u8 y;
      
      b = source[offset++];
      g = source[offset++];
      r = source[offset++];
      
      y = 0.299 * r + 0.587 * g + 0.114 * b;
      dest[dest_offset++] = 0.492 * (float)(b-y);
      dest[dest_offset++] = y;

      b = source[offset++];
      g = source[offset++];
      r = source[offset++];
      
      y = 0.299 * r + 0.587 * g + 0.114 * b;
      dest[dest_offset++] = 0.877 * (float)(r-y);
      dest[dest_offset++] = y;
   }
}

static void rgb322uyvy( __u8 *dest, __u8 *source, int width, int height )
{
   int offset;
   int dest_offset = 0;
   int source_size = width * height * 4;
   
   for( offset = 0; offset < source_size; )
   {
      int r, b, g;
      __u8 y;
      
      r = source[offset++];
      g = source[offset++];
      b = source[offset++];
      offset++;
      
      y = 0.299 * r + 0.587 * g + 0.114 * b;
      dest[dest_offset++] = 0.492 * (float)(b-y);
      dest[dest_offset++] = y;

      r = source[offset++];
      g = source[offset++];
      b = source[offset++];
      offset++;
      
      y = 0.299 * r + 0.587 * g + 0.114 * b;
      dest[dest_offset++] = 0.877 * (float)(r-y);
      dest[dest_offset++] = y;
   }
}

static void rgb24toyuv420p( __u8 *dest, __u8 *source, int width, int height )
{
   int offset = 0;
   int dest_y_offset = 0;
   int dest_u_offset = 0;
   int dest_v_offset = 0;
   int source_size = width * height * 3;
   int x,y;

   __u8 *dest_y;
   __u8 *dest_u;
   __u8 *dest_v;
   
   dest_y = dest;
   dest_u = dest_y + ( width * height );
   dest_v = dest_u + ( width * height ) / 4;
   
   for( y = 0; y < height; y +=2 )
   {
      for( x = 0; x < width; x+= 2 )
      {
	 int r, b, g;
	 __u8 y;
	 
	 r = source[offset++];
	 g = source[offset++];
	 b = source[offset++];
	 
	 y = 0.299 * r + 0.587 * g + 0.114 * b;
	 dest_y[dest_y_offset++] = y;
	 dest_u[dest_u_offset++] = 0.492 * (float)(b-y);
	 dest_v[dest_v_offset++] = 0.877 * (float)(r-y);

	 r = source[offset++];
	 g = source[offset++];
	 b = source[offset++];
	 
	 y = 0.299 * r + 0.587 * g + 0.114 * b;
	 dest_y[dest_y_offset++] = y;
      }
      
      for( x = 0; x < width; x+=2 )
      {
	 int r, b, g;
	 __u8 y;
	 
	 r = source[offset++];
	 g = source[offset++];
	 b = source[offset++];
	 
	 y = 0.299 * r + 0.587 * g + 0.114 * b;
	 dest_y[dest_y_offset++] = y;
	 
	 r = source[offset++];
	 g = source[offset++];
	 b = source[offset++];
	 
	 y = 0.299 * r + 0.587 * g + 0.114 * b;
	 dest_y[dest_y_offset++] = y;
      }
   }
}

static void yuv420ptorgb24( __u8 *dest, __u8 *source, int width, int height )
{
   int x,y;
   __u8 *src_y = source;
   __u8 *src_u = source + width * height;
   __u8 *src_v = src_u + ( width * height ) / 4;


   for( y = 0; y < height; y+=2 )
   {
      for( x = 0; x < width; x+= 2 )
      {
	 __u8 y1, y2, u, v;
	 int c1, c2, d, dg, db, e, er, eg;
	 int ir, ig, ib;
		
	 y1 = src_y[ y * width + x ];
	 y2 = src_y[ y * width + x + 1 ];
	 u = src_u[ ( ( y * width ) / 4 ) + x / 4 ];
	 v = src_v[ ( ( y * width ) / 4 ) + x / 4 ];
      
	 c1 = (y1-16)*298;
	 c2 = (y2-16)*298;
	 d = u-128;
	 dg = 100 * d;
	 db = 516 * d;
	 e = v-128;
	 er = 409 * e;
	 eg = 208 * e;

	 ir = (c1 + er + 128)>>8;
	 ig = (c1 - dg - eg + 128 )>>8;
	 ib = (c1 + db)>>8;

	 *dest++ = (__u8) ( ir > 255 ? 255 : ( ir < 0 ? 0 : ir ) );
	 *dest++ = (__u8) ( ig > 255 ? 255 : ( ig < 0 ? 0 : ig ) );
	 *dest++ = (__u8) ( ib > 255 ? 255 : ( ib < 0 ? 0 : ib ) );
		
	 ir = (c2 + er + 128)>>8;
	 ig = (c2 - dg - eg + 128 )>>8;
	 ib = (c2 + db)>>8;

	 *dest++ = (__u8) ( ir > 255 ? 255 : ( ir < 0 ? 0 : ir ) );
	 *dest++ = (__u8) ( ig > 255 ? 255 : ( ig < 0 ? 0 : ig ) );
	 *dest++ = (__u8) ( ib > 255 ? 255 : ( ib < 0 ? 0 : ib ) );
      }
   }
   
}

static void bgr24torgb24( __u8 *dest, __u8 *source, int width, int height )
{
   int size = width * height;
   int i;
   
   for( i = 0; i < size; i++ )
   {
      unsigned char r, g, b;
      b = *source++;
      g = *source++;
      r = *source++;
      
      *dest++ = r;
      *dest++ = g;
      *dest++ = b;
   }
}

static void rgb32torgb24( __u8 *dest, __u8 *source, int width, int height )
{
   int size = width * height;
   int i;
   

   for( i = 0; i < size; i++ )
   {
      unsigned char r,g,b;
      
      r = *source++;
      g = *source++;
      b = *source++;
      source++;
      
      *dest++ = r;
      *dest++ = g;
      *dest++ = b;
   }
}

static void yuy2toyuyv( __u8 *dest, __u8 *source, int width, int height )
{
   memcpy( dest, source, width * height * 2 );
}




static void by8torgb24( __u8 *dest, __u8* source, int width, int height )
{
   int i,j;
   int r, g, b;
   int yuv;
   int dest_offset = 0;      
   
   for( j = 2; j < height - 2; j+=2 )
   {
      int lineoffset = j * width;
      dest_offset = j * width * 3 + 3;

      for( i = 1; i < width - 2; )
      {
	 int r, g, b;
	 int g1 = source[ lineoffset + i - 1 ];
	 int g2 = source[ lineoffset + i + 1 ];
	 int g3 = source[ ( lineoffset + i ) - width ];
	 int g4 = source[ lineoffset + i + width ];
	 
	 int d1 = ABS( g1 - g2 );
	 int d2 = ABS( g3 - g4 );

	 __u8 y;

	 b = source[ lineoffset + i ];
	 r = ( ( (int)source[ ( ( lineoffset + i ) - width ) + 1 ] + 
		 (int)source[ ( lineoffset + i + width ) + 1 ] + 
		 (int)source[ ( ( lineoffset + i ) - width ) - 1 ] + 
		 (int)source[ ( lineoffset + i + width ) - 1 ] ) / 4 );
	 
	 if( d1 < d2 )
	 {
	    g = ( g1 + g2 ) / 2;
	 }
	 else
	 {
	    g = ( g3 + g4 ) / 2;
	 }	    

	 dest[dest_offset++] = r;
	 dest[dest_offset++] = g;
	 dest[dest_offset++] = b;
	 
	 i++;
	 
	 g = g2;
	 b = ( (int)source[ ( lineoffset + i ) - 1 ] + (int)source[ ( lineoffset + i ) + 1 ] ) / 2;
	 r = ( (int)source[ ( lineoffset + i ) - width ] + (int)source[ lineoffset + i + width ] ) / 2;

	 dest[dest_offset++] = r;
	 dest[dest_offset++] = g;
	 dest[dest_offset++] = b;
	 
	 i++;
      }
    
      lineoffset += width;
      dest_offset = (j+1) * width * 3 + 6;
      
      for( i = 2; i < width - 3; )
      {
	 int r, g, b;
	 int g1 = source[ lineoffset + i - 1 ];
	 int g2 = source[ lineoffset + i + 1 ];
	 int g3 = source[ ( lineoffset + i ) - width ];
	 int g4 = source[ lineoffset + i + width ];
	 
	 int d1 = ABS( g1 - g2 );
	 int d2 = ABS( g3 - g4 );

	 __u8 y;

	 r = source[ lineoffset + i ];

	 if( d1 < d2 )
	 {
	    g = ( g1 + g2 ) / 2;
	 }
	    
	 else
	 {
	    g = ( g3 + g4 ) / 2;
	 }

	 b = ( ( (int)source[ ( ( lineoffset + i ) - width ) + 1 ] + 
		 (int)source[ ( lineoffset + i + width ) + 1 ] + 
		 (int)source[ ( ( lineoffset + i ) - width ) - 1 ] + 
		 (int)source[ ( lineoffset + i + width ) - 1 ] ) / 4 );
	 

	 dest[dest_offset++] = r;
	 dest[dest_offset++] = g;
	 dest[dest_offset++] = b;
	 
	 i++;
	 
	 g = g2;
	 r = ( (int)source[ ( lineoffset + i ) - 1 ] + (int)source[ ( lineoffset + i ) + 1 ] ) / 2;
	 b = ( (int)source[ ( lineoffset + i ) - width ] + (int)source[ lineoffset + i + width ] ) / 2;
	 
	 dest[dest_offset++] = r;
	 dest[dest_offset++] = g;
	 dest[dest_offset++] = b;

	 i++; 
      }
   }
   
}

static void by8tobgr24( __u8 *dest, __u8* source, int width, int height )
{
   int i,j;
   int r, g, b;
   int yuv;
   int dest_offset = 0;      
   
   for( j = 2; j < height - 2; j+=2 )
   {
      int lineoffset = j * width;
      dest_offset = j * width * 3 + 3;

      for( i = 1; i < width - 2; )
      {
	 int r, g, b;
	 int g1 = source[ lineoffset + i - 1 ];
	 int g2 = source[ lineoffset + i + 1 ];
	 int g3 = source[ ( lineoffset + i ) - width ];
	 int g4 = source[ lineoffset + i + width ];
	 
	 int d1 = ABS( g1 - g2 );
	 int d2 = ABS( g3 - g4 );

	 __u8 y;

	 b = source[ lineoffset + i ];
	 r = ( ( (int)source[ ( ( lineoffset + i ) - width ) + 1 ] + 
		 (int)source[ ( lineoffset + i + width ) + 1 ] + 
		 (int)source[ ( ( lineoffset + i ) - width ) - 1 ] + 
		 (int)source[ ( lineoffset + i + width ) - 1 ] ) / 4 );
	 
	 if( d1 < d2 )
	 {
	    g = ( g1 + g2 ) / 2;
	 }
	 else
	 {
	    g = ( g3 + g4 ) / 2;
	 }	    

	 dest[dest_offset++] = b;
	 dest[dest_offset++] = g;
	 dest[dest_offset++] = r;
	 
	 i++;
	 
	 g = g2;
	 b = ( (int)source[ ( lineoffset + i ) - 1 ] + (int)source[ ( lineoffset + i ) + 1 ] ) / 2;
	 r = ( (int)source[ ( lineoffset + i ) - width ] + (int)source[ lineoffset + i + width ] ) / 2;

	 dest[dest_offset++] = b;
	 dest[dest_offset++] = g;
	 dest[dest_offset++] = r;
	 
	 i++;
      }
    
      lineoffset += width;
      dest_offset = (j+1) * width * 3 + 6;
      
      for( i = 2; i < width - 3; )
      {
	 int r, g, b;
	 int g1 = source[ lineoffset + i - 1 ];
	 int g2 = source[ lineoffset + i + 1 ];
	 int g3 = source[ ( lineoffset + i ) - width ];
	 int g4 = source[ lineoffset + i + width ];
	 
	 int d1 = ABS( g1 - g2 );
	 int d2 = ABS( g3 - g4 );

	 __u8 y;

	 r = source[ lineoffset + i ];

	 if( d1 < d2 )
	 {
	    g = ( g1 + g2 ) / 2;
	 }
	    
	 else
	 {
	    g = ( g3 + g4 ) / 2;
	 }

	 b = ( ( (int)source[ ( ( lineoffset + i ) - width ) + 1 ] + 
		 (int)source[ ( lineoffset + i + width ) + 1 ] + 
		 (int)source[ ( ( lineoffset + i ) - width ) - 1 ] + 
		 (int)source[ ( lineoffset + i + width ) - 1 ] ) / 4 );
	 

	 dest[dest_offset++] = b;
	 dest[dest_offset++] = g;
	 dest[dest_offset++] = r;
	 
	 i++;
	 
	 g = g2;
	 r = ( (int)source[ ( lineoffset + i ) - 1 ] + (int)source[ ( lineoffset + i ) + 1 ] ) / 2;
	 b = ( (int)source[ ( lineoffset + i ) - width ] + (int)source[ lineoffset + i + width ] ) / 2;
	 
	 dest[dest_offset++] = b;
	 dest[dest_offset++] = g;
	 dest[dest_offset++] = r;

	 i++; 
      }
   }   
}

static void by8torgb32( __u8 *dest, __u8* source, int width, int height )
{
   int i,j;
   int r, g, b;
   int yuv;
   int dest_offset = 0;      
   
   for( j = 2; j < height - 2; j+=2 )
   {
      int lineoffset = j * width;
      dest_offset = j * width * 3 + 3;

      for( i = 1; i < width - 2; i += 2)
      {
	 int r, g, b;
	 int g1;
	 int g2;
	 int g3;
	 int g4;
	 int d1;
	 int d2;

	 int li = lineoffset + i;

	 __u8 y;

	 g1 = source[ li - 1 ];
	 g2 = source[ li + 1 ];
	 g3 = source[ li - width ];
	 g4 = source[ li + width ];	 
	 d1 = ABS( g1 - g2 );
	 d2 = ABS( g3 - g4 );

	 b = source[ li ];
	 r = ( ( (int)source[ ( li - width ) + 1 ] + 
		 (int)source[ ( li + width ) + 1 ] + 
		 (int)source[ ( li - width ) - 1 ] + 
		 (int)source[ ( li + width ) - 1 ] ) / 4 );
	 
	 if( d1 < d2 )
	 {
	    g = ( g1 + g2 ) / 2;
	 }
	 else
	 {
	    g = ( g3 + g4 ) / 2;
	 }	    

	 dest[dest_offset++] = r;
	 dest[dest_offset++] = g;
	 dest[dest_offset++] = b;
	 dest[dest_offset++] = 0xff;
	 
	 li++;
	 
	 g = g2;
	 b = ( (int)source[ li - 1 ] + (int)source[ li + 1 ] ) / 2;
	 r = ( (int)source[ li - width ] + (int)source[ li + width ] ) / 2;

	 dest[dest_offset++] = r;
	 dest[dest_offset++] = g;
	 dest[dest_offset++] = b;
	 dest[dest_offset++] = 0xff;
	 
      }
    
      lineoffset += width;
      dest_offset = (j+1) * width * 3 + 6;
      
      for( i = 2; i < width - 3; i += 2 )
      {
	 int r, g, b;
	 int g1;
	 int g2;
	 int g3;
	 int g4;
	 int d1;
	 int d2;

	 int li = lineoffset + i;

	 __u8 y;

	 g1 = source[ li - 1 ];
	 g2 = source[ li + 1 ];
	 g3 = source[ li - width ];
	 g4 = source[ li + width ];
	 d1 = ABS( g1 - g2 );
	 d2 = ABS( g3 - g4 );

	 r = source[ li ];

	 if( d1 < d2 )
	 {
	    g = ( g1 + g2 ) / 2;
	 }
	    
	 else
	 {
	    g = ( g3 + g4 ) / 2;
	 }

	 b = ( ( (int)source[ ( li - width ) + 1 ] + 
		 (int)source[ li + width + 1 ] + 
		 (int)source[ ( li - width ) - 1 ] + 
		 (int)source[ ( li + width ) - 1 ] ) / 4 );
	 

	 dest[dest_offset++] = r;
	 dest[dest_offset++] = g;
	 dest[dest_offset++] = b;
	 dest[dest_offset++] = 0xff;
	 
	 li++;
	 
	 g = g2;
	 r = ( (int)source[ li - 1 ] + (int)source[ li + 1 ] ) / 2;
	 b = ( (int)source[ li - width ] + (int)source[ li + width ] ) / 2;
	 
	 dest[dest_offset++] = r;
	 dest[dest_offset++] = g;
	 dest[dest_offset++] = b;
	 dest[dest_offset++] = 0xff;

      }
   }
   
}

static void by8touyvy( __u8 *dest, __u8* source, int width, int height )
{
   int i,j;
   int r, g, b;
   int yuv;
   int dest_offset = 0;      
   
   for( j = 2; j < height - 2; j+=2 )
   {
      int lineoffset = j * width;
      dest_offset = j * width * 2 + 2;

      for( i = 1; i < width - 2; i += 2)
      {
	 int r, g, b;
	 int g1;
	 int g2;
	 int g3;
	 int g4;
	 int d1;
	 int d2;

	 int li = lineoffset + i;

	 __u8 y,u,v;

	 g1 = source[ li - 1 ];
	 g2 = source[ li + 1 ];
	 g3 = source[ li - width ];
	 g4 = source[ li + width ];	 
	 d1 = ABS( g1 - g2 );
	 d2 = ABS( g3 - g4 );

	 b = source[ li ];
	 r = ( ( (int)source[ ( li - width ) + 1 ] + 
		 (int)source[ ( li + width ) + 1 ] + 
		 (int)source[ ( li - width ) - 1 ] + 
		 (int)source[ ( li + width ) - 1 ] ) / 4 );
	 
	 if( d1 < d2 )
	 {
	    g = ( g1 + g2 ) / 2;
	 }
	 else
	 {
	    g = ( g3 + g4 ) / 2;
	 }	    

	 y = 0.299 * (float)r + 0.587 * (float)g + 0.114 * (float)b;
	 u = 0.492 * (float)(b-y);
	 dest[dest_offset++] = u;
	 dest[dest_offset++] = y;
	 
	 li++;
	 
	 g = g2;
	 b = ( (int)source[ li - 1 ] + (int)source[ li + 1 ] ) / 2;
	 r = ( (int)source[ li - width ] + (int)source[ li + width ] ) / 2;

	 y = 0.299 * (float)r + 0.587 * (float)g + 0.114 * (float)b;
	 v = 0.877 * (float)(r-y);
	 dest[dest_offset++] = v;
	 dest[dest_offset++] = y;
	 
      }
    
      lineoffset += width;
      dest_offset = (j+1) * width * 2 + 4;
      
      for( i = 2; i < width - 3; i += 2 )
      {
	 int r, g, b;
	 int g1;
	 int g2;
	 int g3;
	 int g4;
	 int d1;
	 int d2;

	 int li = lineoffset + i;

	 __u8 y,u,v;

	 g1 = source[ li - 1 ];
	 g2 = source[ li + 1 ];
	 g3 = source[ li - width ];
	 g4 = source[ li + width ];
	 d1 = ABS( g1 - g2 );
	 d2 = ABS( g3 - g4 );

	 r = source[ li ];

	 if( d1 < d2 )
	 {
	    g = ( g1 + g2 ) / 2;
	 }
	    
	 else
	 {
	    g = ( g3 + g4 ) / 2;
	 }

	 b = ( ( (int)source[ ( li - width ) + 1 ] + 
		 (int)source[ li + width + 1 ] + 
		 (int)source[ ( li - width ) - 1 ] + 
		 (int)source[ ( li + width ) - 1 ] ) / 4 );
	 

	 y = 0.299 * (float)r + 0.587 * (float)g + 0.114 * (float)b;
	 u = 0.492 * (float)(b-y);
	 dest[dest_offset++] = u;
	 dest[dest_offset++] = y;
	 
	 li++;
	 
	 g = g2;
	 r = ( (int)source[ li - 1 ] + (int)source[ li + 1 ] ) / 2;
	 b = ( (int)source[ li - width ] + (int)source[ li + width ] ) / 2;
	 
	 y = 0.299 * (float)r + 0.587 * (float)g + 0.114 * (float)b;
	 v = 0.877 * (float)(r-y);
	 dest[dest_offset++] = v;
	 dest[dest_offset++] = y;

      }
   }
   
}
